import re
from datetime import datetime
from string import Template

import writer
from cursor_provider import CursorProvider


class Query:

    def __init__(self, source, template_vars=None):
        self.source = source
        self.template_vars = template_vars if template_vars is not None else dict()
        self.__interpolated = None
        self.__query_str = None
        self.__result = None

    @property
    def interpolated(self):
        if self.__interpolated is not None:
            return self.__interpolated
        else:
            template = Template(open(self.source).read())
            interpolated = template.safe_substitute(**self.template_vars)
            unprovided_filtered = re.sub('\n.*\\${\\w+}.*\n', '\n', interpolated)
            self.__interpolated = unprovided_filtered
            return self.__interpolated

    @property
    def query_str(self):
        if self.__query_str is not None:
            return self.__query_str
        else:
            self.__query_str = ' '.join([
                normalize_space(strip_inline_comment(line).strip())
                for line in self.interpolated.split('\n')
                if not is_comment(line) and not is_blank(line)
            ])
            return self.__query_str

    @property
    def result(self):
        if self.__result is not None:
            return self.__result
        else:
            self.__result = Result(self.query_str)
            return self.__result

    def export(self, destination=None):
        return writer.write(self, destination)


def normalize_space(line):
    return re.sub(' +', ' ', line)


def strip_inline_comment(line):
    return re.sub('(--|#).*', '', line)


def is_blank(line):
    return not line.strip()


def is_comment(line):
    return line.strip().startswith('--') or line.strip().startswith('#')


class Result:

    def __init__(self, query_str):
        cursor = CursorProvider.cursor()
        self.execution_start = datetime.now()
        cursor.execute(query_str)
        self.execution_end = datetime.now()
        self.duration = self.execution_end - self.execution_start
        self.rows = cursor.fetchall()
        self.description = tuple(column[0] for column in cursor.description)



from getpass import getpass

import keyring
from alone import MetaSingleton
from mysql import connector

import chest


class CursorProvider(metaclass=MetaSingleton):

    def __init__(self):
        service = 'CursorProvider-{host}'.format(host=chest.host, db=chest.database)
        keyring_password = keyring.get_password(service, chest.user)
        if keyring_password is None or chest.ask_password:
            password = getpass("Password for {user}@{host}: ".format(user=chest.user, host=chest.host))
        else:
            password = keyring_password

        self.connection = connector.connect(
            host=chest.host,
            user=chest.user,
            db=chest.database,
            passwd=password)

        if password is not keyring_password and chest.store_password:
            keyring.set_password(service, chest.user, password)

    def __del__(self):
        if hasattr(self, 'connection') and self.connection.is_connected():
            self.connection.close()

    @staticmethod
    def cursor():
        return CursorProvider().connection.cursor()



import argparse

import chest
from query import Query

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('query', nargs='+', help='Path to a file containing a single sql statement')
    parser.add_argument('--host', required=True, help='MySQL server host')
    parser.add_argument('--user', required=True, help='MySQL server user')
    parser.add_argument('--database', help='MySQL database name')
    parser.add_argument('-a', '--ask-password', default=False, action='store_true',
                        help='Do not try to retrieve password from keyring, always ask password')
    parser.add_argument('-s', '--store_password', default=False, action='store_true',
                        help='Store password into keyring after connecting to the database')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-d', '--destination', help='Directory where to export results')
    group.add_argument('--display', default=False, action='store_true',
                       help='Do not export results but display them to stdout')
    args = parser.parse_args()

    chest.host = args.host
    chest.user = args.user
    chest.database = args.database
    chest.ask_password = args.ask_password
    chest.store_password = args.store_password

    queries = [Query(path) for path in args.query]
    for query in queries:
        if not args.display:
            query.export(destination=args.destination)
        else:
            print(query.source)
            print(query.result.description)
            print(query.result.rows)



import csv
from os.path import join, dirname, basename, splitext
from shutil import copyfile


REPORT_TEMPLATE = '''
-- START TIME: {start}
-- END TIME: {end}
-- DURATION: {duration}
-- ROWS COUNT: {count}
-- RESULT FILE: {file}
'''


def write(query, destination=None):
    prefix = query.result.execution_start.strftime("%Y-%m-%dT%H-%M-%S_")

    directory = destination if destination is not None else dirname(query.source)
    report_path = join(directory, prefix + basename(query.source))
    export_path = splitext(report_path)[0] + '.csv'

    copyfile(query.source, report_path)

    with open(report_path, 'a') as report_file:
        report_file.write(REPORT_TEMPLATE.format(
            start=query.result.execution_start.isoformat(),
            end=query.result.execution_end.isoformat(),
            duration=query.result.duration,
            count=len(query.result.rows),
            file=basename(export_path) if len(query.result.rows) > 0 else None
        ))

    if len(query.result.rows) > 0:
        with open(export_path, 'w') as export_file:
            csv_writer = csv.writer(export_file, quoting=csv.QUOTE_ALL)
            csv_writer.writerow(query.result.description)
            for row in query.result.rows:
                csv_writer.writerow(row)

    return report_path, export_path
